/* "eqcs_state_rep.cc"
 * Copyright (C) 1999 Peter Belkner <pbelkner@berlin.snafu.de>
 *
 * This file is part of the eqcs library. This library is free
 * software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License as published by the Free Software
 * Foundation; either version 2, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; see the file COPYING.  If not, write to the
 * Free Software Foundation, 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
#ifdef __GNUG__
#pragma implementation
#endif

#include <assert.h>
#include <iostream>
#include <cstdlib>
#include "eqcs_sep.h"
#include "eqcs_state_rep.h"

/*****************************************************************************\
*	state representation													  *
\*****************************************************************************/

EqcsStateRep::EqcsStateRep(int nqubits, int)
	{ _nqubits = nqubits; }

EqcsStateRep::EqcsStateRep(unsigned long bits)
{
	bitset_t bitset = bitset_t::bits2bitset(bits);
	_nqubits = bitset.nbits();
	_terms[bitset] = 1.0;
}

EqcsStateRep::EqcsStateRep(int nqubits, const terms_t &map) : _terms(map)
	{ _nqubits = nqubits; }

EqcsStateRep::~EqcsStateRep()
	{}

bool EqcsStateRep::less_then(const EqcsStateRep *rep) const
{
	double n1 = norm(this);
	double n2 = norm(rep);

	terms_t::const_reverse_iterator i1 = _terms.rbegin();
	terms_t::const_reverse_iterator i2 = rep->_terms.rbegin();

	for (;;) {
		// Skip the left hand side's vanishing terms.
		while (i1 != _terms.rend() && i1->second == 0.0)
			++i1;

		// Skip the right hand side's vanishing terms.
		while (i2 != rep->_terms.rend() && i2->second == 0.0)
			++i2;

		// Both sides contain the same non-vanishing terms.
		if (i1 == _terms.rend() && i2 == rep->_terms.rend())
			return false;

		// The right hand side contains at least one more non-vanishing term.
		if (i1 == _terms.rend())
			return true;

		// The left hand side contains at least one more non-vanishing term.
		if (i2 == rep->_terms.rend())
			return false;

		// Compare the base vectors.
		if (i1->first < i2->first)
			return true;
		if (i1->first > i2->first)
			return false;

		complex_t c1 = i1->second / n1;
		complex_t c2 = i2->second / n2;

		// Compare the coefficient's real parts.
		if (c1.real() < c2.real())
			return true;
		if (c1.real() > c2.real())
			return false;

		// Compare the coefficient's imaginary parts.
		if (c1.imag() < c2.imag())
			return true;
		if (c1.imag() > c2.imag())
			return false;

		++i1;
		++i2;
	}

	// We should never go here.
	assert(false);

	return false;
}

EqcsStateRep *EqcsStateRep::neg() const
{
	EqcsStateRep *state = new EqcsStateRep(_nqubits, 0);

	for (const_iterator p = _terms.begin(); p != _terms.end(); ++p)
		state->_terms[p->first] = -p->second;

	return state;
}

EqcsStateRep *EqcsStateRep::add(const EqcsStateRep *rep)
{
	if (_nqubits < rep->_nqubits)
		_nqubits = rep->_nqubits;

	for (const_iterator p = rep->_terms.begin(); p != rep->_terms.end(); ++p)
		_terms[p->first] += p->second;

	return this;
}

EqcsStateRep *EqcsStateRep::sub(const EqcsStateRep *rep)
{
	if (_nqubits < rep->_nqubits)
		_nqubits = rep->_nqubits;

	for (const_iterator p = rep->_terms.begin(); p != rep->_terms.end(); ++p)
		_terms[p->first] -= p->second;

	return this;
}

EqcsStateRep *EqcsStateRep::mul(const complex_t &c)
{
	for (iterator p = _terms.begin(); p != _terms.end(); ++p)
		p->second *= c;

	return this;
}

/* A generalization from Rafal's first and second collapsing
 * algorithms.
 *
 * Rafal's second collapsing algorithm may be viewed as collapsing
 * a set of one qubit. Rafal's first collapsing algorithm may be
 * viewed as collapsing the set of all qubits.
 *
 * The generalized algorithm should be something like this:
 *
 *     - Calculate the partial probabilities for each (sub)
 *       bit pattern defined by the set in question by adding
 *       the expectation values of base states with the same
 *       (sub) bit pattern.
 *     - Coose a (sub) bit pattern by applying Rafal's first
 *       algorithm or Bernhard's algorithm to the sequence of
 *       partial probabilities.
 *     - Project the state into a subspace by dropping all
 *       terms not containing the choosen (sub) bit pattern.
 */
EqcsStateRep *EqcsStateRep::collapse(unsigned long bits) const
{
	// Calculate the partial probabilities for each (sub) bit
	// pattern defined by the set in question by adding the
	// expectation values of base states with the same (sub) bit
	// pattern.
	typedef map<bitset_t, double> table_t;

	table_t table;		// A table of partial probabilities.

	bitset_t bitset = bitset_t::bits2bitset(bits);

	for (const_iterator i = _terms.begin(); i != _terms.end(); ++i)
		table[bitset & i->first] += norm(i->second);

	// Coose a (sub) bit pattern by applying Rafal's first algorithm
	// or Bernhard's algorithm to the sequence of partial probabilities.
	double x = 0.0;
	double r = (norm(this) * rand()) / ((double)RAND_MAX + 1.0);
	table_t::const_iterator j = table.begin();

	while ((x += j->second) < r)
		if (table.end() == ++j) {
			// Theoretically we should never go here.
			cerr << "Unexpected error #7891" << endl;
			--j;

			break;
		}

	// Filter out the bits of interest.
	bitset_t mask = j->first & bitset;

	// Project the state into the subspace, i.e. copy all terms
	// with a (sub) bit pattern equal to the mask.
	EqcsStateRep *state = new EqcsStateRep(_nqubits, 0);

	for (const_iterator i = _terms.begin(); i != _terms.end(); ++i)
		if (mask == (i->first & bitset))
			state->_terms[i->first] = i->second;

	return state;
}

// Calculate a quantum state's norm, i.e. |q0|^2 + |q1|^2 + ...
double norm(const EqcsStateRep *state)
{
	typedef EqcsStateRep::terms_t terms_t;
	const terms_t &terms = state->_terms;
	double n = 0.0;

	for (terms_t::const_iterator i = terms.begin(); i != terms.end(); ++i)
		n += norm(i->second);

	return n;
}

ostream &EqcsStateRep::put(ostream &os) const
{
	// Calculate the norm for renormalization.
	double n = sqrt(norm(this));

	EqcsSep sep(" + ");

	// For each state ...
	for (const_iterator p = _terms.begin(); p != _terms.end(); ++p) {
		if (0.0 == p->second)
			continue;

		// Write the term.
		os << sep << (p->second / n) << ket(p->first, _nqubits);
	}

	return os;
}

ostream &EqcsStateRep::put_bits(ostream &os, int nregs, int regwidth) const
{
	// Calculate the norm for renormalization.
	double n = sqrt(norm(this));

	EqcsSep sep(" + ");

	// For each state ...
	for (const_iterator p = _terms.begin(); p != _terms.end(); ++p) {
		if (0.0 == p->second)
			continue;

		// Write the term.
		os << sep << (p->second / n) << bits(p->first, nregs, regwidth);
	}

	return os;
}

ostream &EqcsStateRep::put_nums(ostream &os, int nregs, int regwidth) const
{
	// Calculate the norm for renormalization.
	double n = sqrt(norm(this));

	EqcsSep sep(" + ");

	// For each state ...
	for (const_iterator p = _terms.begin(); p != _terms.end(); ++p) {
		if (0.0 == p->second)
			continue;

		// Write the term.
		os << sep << (p->second / n) << nums(p->first, nregs, regwidth);
	}

	return os;
}
